Considerater = LibStub("AceAddon-3.0"):NewAddon("Considerater", "AceConsole-3.0","AceEvent-3.0")

local statCategories = {
  ["Stats"] = {
    ["STR"] = { desc = "Strength" }, 
    ["AGI"] = { desc = "Agility" }, 
    ["STA"] = { desc = "Stamina" }, 
    ["INT"] = { desc = "Intelligence" }, 
    ["SPI"] = { desc = "Spirit" }, 
  },

  ["Offense"] = {
    ["DPS"] = { desc = "DPS" }, 
    ["AP"] = { desc = "Attack Power" }, 
    ["SPELL_DMG"] = { desc = "Spell Power" }, 
    ["MELEE_HIT_RATING"] = { desc = "Hit Rating" }, 
    ["MELEE_CRIT_RATING"] = { desc = "Crit Rating" }, 
    ["MELEE_HASTE_RATING"] = { desc = "Haste Rating" }, 
    ["EXPERTISE_RATING"] = { desc = "Expertise Rating" }, 
    ["IGNORE_ARMOR"] = { desc = "Ignore Armor" }, 
    ["ARMOR_PENETRATION_RATING"] = { desc = "Armor Pen." }, 
    ["SPELLPEN"] = { desc = "Spell Pen." }, 
    ["MAX_DAMAGE"] = { desc = "Top End Damage" }, 
  },

  ["Defense"] = {
    ["ARMOR"] = { desc = "Armor" }, 
    ["ARMOR_BONUS"] = { desc = "Bonus Armor" }, 
    ["DEFENSE_RATING"] = { desc = "Defense Rating" }, 
    ["DODGE_RATING"] = { desc = "Dodge Rating" }, 
    ["PARRY_RATING"] = { desc = "Parry Rating" }, 
    ["BLOCK_RATING"] = { desc = "Block Rating" }, 
    ["RESILIENCE_RATING"] = { desc = "Resilience Rating" }, 
    ["BLOCK_VALUE"] = { desc = "Block Value" }, 
  },

  ["Regen"] = {
    ["MANA_REG"] = { desc = "MP5" }, 
    ["HEALTH_REG"] = { desc = "HP5" }, 
  },

  ["Resists"] = {
    ["FIRE_RES"] = { desc = "Fire Resist" }, 
    ["NATURE_RES"] = { desc = "Nature Resist" }, 
    ["FROST_RES"] = { desc = "Frost Resist" }, 
    ["SHADOW_RES"] = { desc = "Shadow Resist" }, 
    ["ARCANE_RES"] = { desc = "Arcane Resist" }, 
  },
}

local equipLocs = {
  { value = "INVTYPE_HEAD", text = "Head"},
  { value = "INVTYPE_NECK", text = "Neck"},
  { value = "INVTYPE_SHOULDER", text = "Shoulder"},
  { value = "INVTYPE_CHEST", text = "Chest"},
  { value = "INVTYPE_WAIST", text = "Waist"},
  { value = "INVTYPE_LEGS", text = "Legs"},
  { value = "INVTYPE_FEET", text = "Feet"},
  { value = "INVTYPE_WRIST", text = "Wrist"},
  { value = "INVTYPE_HAND", text = "Hand"},
  { value = "INVTYPE_CLOAK", text = "Cloak"},
  { value = "INVTYPE_FINGER", text = "Finger"},
  { value = "INVTYPE_TRINKET", text = "Trinket"},
  { value = "INVTYPE_WEAPON", text = "Weapon"},
  { value = "INVTYPE_SHIELD", text = "Shield"},
  { value = "INVTYPE_2HWEAPON", text = "2H Weapon"},
  { value = "INVTYPE_HOLDABLE", text = "Off-Hand Item"},
  { value = "INVTYPE_RANGED", text = "Ranged"},
  { value = "INVTYPE_RELIC", text = "Relic"},
}

local equipLocAliases = {
  ["INVTYPE_CHEST"] = { "INVTYPE_CHEST", "INVTYPE_BODY", "INVTYPE_ROBE" },
  ["INVTYPE_WEAPON"] = { "INVTYPE_WEAPON", "INVTYPE_WEAPONMAINHAND", "INVTYPE_WEAPONOFFHAND" },
  ["INVTYPE_RANGED"] = { "INVTYPE_RANGED", "INVTYPE_THROWN" },
}

-- Build edit profile options panel
function Considerater:EditProfileOptions() 
  local opts
  local statOpts = {}

  for cat,stats in pairs(statCategories) do
    local catOpts = {}
    for statName,stat in pairs(stats) do
      catOpts[statName] = {
        type = 'range',
        name = stat["desc"],
        min = 0,
        max = 25,
        bigStep = 0.1,
        step = 0.01,
        get = function (info) 
          local prof = 
            Considerater.db.profile.profiles[Considerater.curEditProfile]
          if not prof or not prof[info[#info]] then
            return 0
          else
            return prof[info[#info]]
          end
        end,
        set = function (info,val) 
          local prof = 
            Considerater.db.profile.profiles[Considerater.curEditProfile]
          if prof and type(val) == "number" then
            prof[info[#info]] = val
          end
        end,
      }
    end

    statOpts[cat] = {
      type = 'group',
      name = cat,
      args = catOpts,
    }
  end
  
  local args = {}
  args["desc"] = {
    type = 'description',
    name = 'Select a profile and adjust the weighting of different stats',
    order = 0,
  }

  args["profileSelect"] = {
    type = 'select',
    name = 'Profile',
    order = 50,
    desc = 'Profile to edit',
    values = function () return Considerater:GetProfileNames() end,
    get = function (info) 
      for i,n in pairs(Considerater:GetProfileNames()) do
        if n == Considerater.curEditProfile then
          return i
        end
      end
      return 0
    end,
    set = function (info,key) 
      if type(key) == "number" then
        key = Considerater:GetProfileNames()[key]
      end
      Considerater.curEditProfile = key
    end,
  }

  for a,b in pairs(statOpts) do
    args[a] = b
  end
    
  args["ResetGroup"] = {
    type = 'group',
    name = 'Reset',
    order = -1,
    args = {
      ["ResetDesc"] = {
        type = 'description',
        name = 'Reset profile to defaults',
        width = 'half',
        order = 0,
      },
      ["Reset"] = {
        type = 'execute',
        name = 'Reset',
        width = 'half',
        order = -1,
        func = function (info) 
          self:Print("Reseting Profile: "..Considerater.curEditProfile)
          Considerater:ResetProfile(Considerater.curEditProfile) 
        end,
      }
    },
  }

  opts = {
    type = 'group',
    name = 'Edit Profiles',
    args = args,
  }
  return opts
end

local newProfileOptions = {
  name = 'New Profile',
  handler = Considerater,
  type = 'group',
  args = {
    desc = {
      type = 'description',
      name = 'Type a new name in, press Okay, and then press Create',
      order = 0,
    },
    input = {
      type = 'input',
      name = 'New Profile Name',
      order = 100,
      get = function (info) return Considerater.newProfileName end,
      set = function (info,val) Considerater.newProfileName = val end,
    },
    create = {
      type = 'execute',
      name = 'Create',
      order = 150,
      width = 'half',
      func = function (info) 
        local name = Considerater.newProfileName
        if name and #name > 0 then
          for pname, p in pairs(Considerater.db.profile.profiles) do
            if pname == name then
              Considerater:Print("Profile by that name already exists.")
              return
            end
          end
          Considerater:Print("Creating profile: "..name)
          Considerater.db.profile.profiles[name] = {}
        end
      end,
    },
  },
}

local delProfileOptions = {
  name = 'Delete Profile',
  handler = Considerater,
  type = 'group',
  args = {
    desc = {
      type = 'description',
      name = 'Select a profile from the list and then press Delete',
      order = 0,
    },
    input = {
      type = 'select',
      name = 'Profile',
      desc = 'Profile to delete',
      order = 100,
      values = function () return Considerater:GetProfileNames() end,
      get = function (info) 
        for i,n in pairs(Considerater:GetProfileNames()) do
          if n == Considerater.curDelProfile then
            return i
          end
        end
        return 0
      end,
      set = function (info,key) 
        if type(key) == "number" then
          key = Considerater:GetProfileNames()[key]
        end
        Considerater.curDelProfile = key
      end,
    },
    button = {
      type = 'execute',
      name = 'Delete',
      order = 150,
      func = function(info) 
        if Considerater.curDelProfile then
          Considerater.db.profile.profiles[Considerater.curDelProfile] = nil
        end
      end
    },
    nb = {
      type = 'description',
      name = '(Note that pressing Cancel afterwards will revert any profile changes)',
      order = 200,
    },
  },
}

local options = {
  name = "Considerater",
  handler = Considerater,
  type = 'group',
  args = {
    desc = {
      type = 'description',
      name = 'Select all profiles of interest to use when rating items.',
      order = 0,
    },
    activeProfiles = {
      type = 'multiselect',
      name = "Active Profiles",
      desc = "Stat profiles to score items against",
      values = function () return Considerater:GetProfileNames() end,
      get = function (info,key) 
        if type(key) == "number" then
          key = Considerater:GetProfileNames()[key]
        end
        return Considerater:IsProfileActive(key) 
      end,
      set = function (info,key,val) 
        if type(key) == "number" then
          key = Considerater:GetProfileNames()[key]
        end
        Considerater:SetProfileActive(key,val) 
      end,
    },
  },
}

local defaultRestoDruidProfile = {
  ["SPI"] = 8,
  ["INT"] = 6,
  ["STA"] = 4,
  ["SPELL_DMG"] = 18.8,
  ["MANA_REG"] = 24,
}

local defaultCatDruidProfile = {
  ["STR"] = 2.266,
  ["AGI"] = 2.812,
  ["MELEE_HASTE_RATING"] = 1.443,
  ["MELEE_CRIT_RATING"] = 1.676,
  ["MELEE_HIT_RATING"] = 2.07,
  ["AP"] = 1,
  ["FERAL_AP"] = 1,
  ["ARMOR_PENETRATION_RATING"] = 2.247,
  ["EXPERTISE_RATING"] = 2.096,
}

local defaultBearDruidProfile = {
  ["STA"] = 4.33,
  ["AGI"] = 6.79,
  ["ARMOR"] = 0.59,
  ["DEFENSE_RATING"] = 2.79,
  ["DODGE_RATING"] = 4.43,
  ["EXPERTISE_RATING"] = 1.69,
}

local defaultBalanceDruidProfile = {
  ["SPELL_CRIT_RATING"] = 15,
  ["INT"] = 6,
  ["STA"] = 4,
  ["SPELL_DMG"] = 18.8,
  ["MANA_REG"] = 10,
}

local defaultProfiles = { 
  ["Resto Druid"] = defaultRestoDruidProfile,
  ["Cat Druid"] = defaultCatDruidProfile,
  ["Bear Druid"] = defaultBearDruidProfile,
  ["Balance Druid"] = defaultBalanceDruidProfile,
}

local defaultActiveProfiles = {
  "Resto Druid", 
  "Cat Druid", 
  "Bear Druid",
  "Balance Druid",
}

local defaults = {
  profile = {
    profiles = nil,
  },
  char = {
    cachedItems = {}, 
    activeProfiles = defaultActiveProfiles,
    gemProfiles = {},
  },
}

function Considerater:OnInitialize()
  self.db = LibStub("AceDB-3.0"):New("ConsideraterDB", defaults, "Default")
  if not self.db.profile.profiles then
    self.db.profile.profiles = self:DeepCopy(defaultProfiles)
  end
  LibStub("AceConfig-3.0"):RegisterOptionsTable("Considerater", options)
  LibStub("AceConfig-3.0"):RegisterOptionsTable("Considerater_Edit", self:EditProfileOptions())
  LibStub("AceConfig-3.0"):RegisterOptionsTable("Considerater_New", newProfileOptions)
  LibStub("AceConfig-3.0"):RegisterOptionsTable("Considerater_Del", delProfileOptions)
  self:RegisterChatCommand("consider", "ChatCommand")

  -- Setup Blizzard Interface Options frames
  self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Considerater", "Considerater")
  self.editProfileOptionsFrame =
    LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Considerater_Edit", "Edit Profiles","Considerater")
  self.newProfileOptionsFrame =
    LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Considerater_New", "New Profile","Considerater")
    LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Considerater_Del", "Delete Profile","Considerater")
  -- Callback for Okay button:
  self.editProfileOptionsFrame.okay = function ()
    if not Considerater:DeepEqual(Considerater.db.profile.profiles, Considerater.backupProfiles) then
      Considerater:Print("Profiles were changed.  Please revisit your bank to re-evaluate all your items.")
      Considerater:RefreshGemProfiles()
    end
    Considerater.backupProfiles = nil
  end
  -- Callback for Cancel button:
  self.editProfileOptionsFrame.cancel = function ()
    if Considerater.backupProfiles then
      Considerater.db.profile.profiles = Considerater:DeepCopy(Considerater.backupProfiles)
      Considerater.backupProfiles = nil
    end
  end
  -- Override OnShow handler in order to copy current state
  self.editProfileOptionsFrame:SetScript("OnShow", function(frame)
    if not Considerater.backupProfiles then
      Considerater.curEditProfile = Considerater:GetProfileNames()[1]
      Considerater.backupProfiles = Considerater:DeepCopy(Considerater.db.profile.profiles)
    end
    -- copied from AceGUIWidth-BlizOptionsFrame.lua:
    this.obj:Fire("OnShow") -- do AceGUI OnShow stuff
  end)

  self:RegisterEvent("BANKFRAME_OPENED")
  self:RegisterEvent("BAG_UPDATE")
  self:RegisterEvent("BANKFRAME_CLOSED")
  self:RegisterEvent("UNIT_INVENTORY_CHANGED")
  self:RegisterEvent("START_LOOT_ROLL")
  self:RegisterEvent("LOOT_OPENED")
  
  self.bankIsOpen = false
  self.dirtyBags = {}
end

function Considerater:OnEnable()
end

function Considerater:OnDisable()
end

function Considerater:ChatCommand(input)
  if not input or input:trim() == "" then
    self:ConsiderUI()
  else
    -- LibStub("AceConfigCmd-3.0").HandleCommand(Considerater, "consider", "Considerater", input)
    self:ConsiderCmd(input)
  end
end

function Considerater:ConsiderCmd(arg)
  local cmd = arg:trim()
  if cmd == "refresh" then
    self:RefreshGemProfiles()
    if bankIsOpen then
      self:ClearItemCache()
      self:SaveItems()
    end
    self:ScoreCachedItems()
  elseif cmd == "save" then
    self:SaveItems()
  elseif cmd == "clearcache" then
    self:ClearItemCache()
  elseif cmd == "resetall" then
    self:ResetProfiles()
    self:ClearItemCache()
    self.db.char.gemProfiles = {}
  elseif cmd == "showgems" then
    for p,gems in pairs(self.db.char.gemProfiles) do
      for c,gem in pairs(gems) do
        self:Print(p..": "..c..": "..gem["link"])
      end
    end
  elseif cmd == "edit" then
    InterfaceOptionsFrame_OpenToCategory(self.editProfileOptionsFrame)
  elseif cmd == "new" then
    InterfaceOptionsFrame_OpenToCategory(self.newProfileOptionsFrame)
  elseif cmd == "setup" or cmd == "config" or cmd == "configure" or cmd == "options" then
    InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
  elseif cmd:sub(1,5) == "check" then
    local param = arg:trim():sub(6):trim()
    local sum = StatLogic:GetSum(param)
    if sum then
      local topScores = self:Consider(param)
      for profile,data in pairs(topScores) do
        if self:EqualItemLinks(data["link"], sum["link"]) then
          self:Print(profile..": "..data["link"].." has top score = "..data["score"])
        end
      end
    end
  else
    self:AddOtherItem(arg:trim(), "Linked")
    self:ConsiderUI()
    self:UIRefreshItemList()
  end
end

function Considerater:BAG_UPDATE(event, bag)
  --self:Print("BAG_UPDATE: "..bag)
  -- For some reason, WoW generates spurious events whenever the main
  -- backpack is updated, but with bag = -2,-4, or -5.
  if bag > -2 then
    self:MarkBlizBagDirty(bag)
  end
  -- WoW never generates a BAG_UPDATE event for the main bank container, but,
  -- we workaround that by always caching everything whenever the bank is open.
end

function Considerater:BANKFRAME_OPENED(event)
  --self:Print("BANKFRAME_OPENED")
  self.bankIsOpen = true
  self:ClearItemCache()
  self:SaveItems()
  self:ScoreCachedItems()
end

function Considerater:BANKFRAME_CLOSED(event)
  --self:Print("BANKFRAME_CLOSED: "..GetNumBankSlots())
  self.bankIsOpen = false
end

function Considerater:UNIT_INVENTORY_CHANGED(event, arg1)
  --self:Print("UNIT_INVENTORY_CHANGED: "..arg1)
  self:MarkInvDirty()
end

function Considerater:LOOT_OPENED(event, arg1)
  if self.debug then
    self:Print("LOOT_OPENED: "..arg1)
    self:Print("From: "..UnitName("target"))
    self:Print("GetNumLootItems: "..GetNumLootItems())
  end
  for i = 1, GetNumLootItems() do
    if (LootSlotIsItem(i)) then
      local link = GetLootSlotLink(i)
      if self.debug then
        self:Print("Item "..i..": "..link)
      end
      self:AddOtherItem(link, "Loot")
    end
  end
end

function Considerater:START_LOOT_ROLL(event, arg1, arg2)
  if self.debug then
    self:Print("START_LOOT_ROLL: "..arg1.." "..arg2)
    self:Print("Rolling on: "..GetLootRollItemLink(arg1))
  end
  self:AddOtherItem(GetLootRollItemLink(arg1), "Loot")
end

function Considerater:AddOtherItem(link, origin)
  local equipLoc = self:GetItemEquipLoc(link)

  if equipLoc and #equipLoc > 0 then
    if not self.otherItemsList then
      self.otherItemsList = {}
    end
    if not self.otherItemsList[link] then
      self.otherItemsList[link] = {}
    end
    
    self.otherItemsList[link]["link"] = link
    self.otherItemsList[link]["loc"] = equipLoc
    self.otherItemsList[link]["origin"] = origin
    self.otherItemsList[link]["scores"] = {}
    for _,prof in pairs(self.db.char.activeProfiles) do
      self.otherItemsList[link]["scores"][prof] =
        self:ScoreItem(link, prof)
    end
  end
end

function Considerater:OnUpdate()
  getglobal("ConsideraterUpdateFrame"):Hide() -- disable OnUpdate for now
  for i,v in pairs(self.dirtyBags) do
    if v then
      if bankIsOpen then
        self:ClearItemCache()
        self:SaveItems()
        self:ScoreCachedItems()
        return
      else
        -- Do not refresh bank bags if bank is not open
        if bankIsOpen or i == 0 or not (i >= 7 and i <= 13) then
          self:ClearItemCacheInBag(i)
          self:SaveItemsInBag(i)
          self:ScoreCachedItemsInBag(i)
        end
      end
    end
  end
end

function Considerater:ScoreSummary(sumTable, profile)
  local profTable = self.db.profile.profiles[profile]
  local sum = 0
  if sumTable and profTable then
    for k,v in pairs(sumTable) do
      if type(v) == "number" then
        local w = profTable[k]
        if w then
          sum = sum + v * w
        end
      end
    end
  end
  return sum
end

-- Score an item (link) by a profile (name)
function Considerater:ScoreItem(item, profile)
  local sumTable = self:GetNormalizedSum(item, profile)
  return self:ScoreSummary(sumTable, profile)
end

function Considerater:Consider(item) 
  local topScores = {}
  local loc = self:GetItemEquipLoc(item)

  -- If not wearable, then no result
  if not loc or loc == "" then
    return {}
  end

  -- Gem logic: if item has gems then must re-gem for each profile.
  local strippedItem = StatLogic:RemoveEnchantGem(item)
  local score
  local gemmed = true
  local sum = StatLogic:GetSum(strippedItem)

  if (sum["EMPTY_SOCKET_RED"] or
      sum["EMPTY_SOCKET_YELLOW"] or
      sum["EMPTY_SOCKET_BLUE"] or
      sum["EMPTY_SOCKET_META"]) then
      gemmed = true
  end

  item = strippedItem

  -- Compute scores for each active profile
  for i,profile in pairs(self.db.char.activeProfiles) do
    if gemmed then
      item = self:NormalizeGems(strippedItem,profile)
      sum = StatLogic:GetSum(item)
    end
    score = self:ScoreSummary(sum,profile)
    topScores[profile] = {
      ["score"] = score,     -- may be modified below
      ["link"] = item,
      ["itemScore"] = score, -- won't be modified below
    }
  end

  -- Now compare against all items cached
  for i,bag in pairs(self.db.char.cachedItems) do
    for j,slot in pairs(bag) do
      if slot["loc"] == loc then
        for profile,score in pairs(slot["scores"]) do
          if score > topScores[profile]["score"] then
            -- new top score found
            topScores[profile]["score"] = score
            topScores[profile]["link"] = slot["link"]
          end
        end
      end
    end
  end
  return topScores
end

function Considerater:ScoreCachedItemsInBag(bag)
  for j,slot in pairs(self.db.char.cachedItems[bag]) do
    local strippedItem = StatLogic:RemoveEnchantGem(slot["link"])
    local sum = StatLogic:GetSum(strippedItem)
    local gemmed = false
    if (sum["EMPTY_SOCKET_RED"] or
        sum["EMPTY_SOCKET_YELLOW"] or
        sum["EMPTY_SOCKET_BLUE"] or
        sum["EMPTY_SOCKET_META"]) then
        gemmed = true
    end
    for k,profile in pairs(self.db.char.activeProfiles) do
      if gemmed then
        sum = StatLogic:GetSum(self:NormalizeGems(strippedItem, profile))
      end
      slot["scores"][profile] = self:ScoreSummary(sum,profile)
    end
  end
end

function Considerater:ScoreCachedItems()
  for i,bag in pairs(self.db.char.cachedItems) do
    self:ScoreCachedItemsInBag(i)
  end
end

function Considerater:ClearItemCacheInBag(bag)
  self.db.char.cachedItems[bag] = {}
end

function Considerater:ClearItemCache()
  self.db.char.cachedItems = {}
end

-- use "virtual bag indexing" based around the helper functions at the bottom
-- of this file
function Considerater:SaveItemsInBag(bag)
  if not self.db.char.cachedItems then
    self.db.char.cachedItems = {}
  end
  local numSlots = self:GetMaxSlotNum(bag)
  if numSlots and numSlots > 0 then
    -- Only cache bags which are openable at this time
    for slot = 1,numSlots do
      local item = self:GetBagSlotItemLink(bag,slot)
      if item then
        local loc = self:GetItemEquipLoc(item)
        if loc and loc ~= "" then
          if not self.db.char.cachedItems[bag] then
            -- Create table for bag if it does not exist
            self.db.char.cachedItems[bag] = {}
          end
          -- Save it
          self.db.char.cachedItems[bag][slot] = { 
            ["link"] = item,
            ["loc"] = loc,
            ["scores"] = {},
          }
        end
      end
    end
  end
  self.dirtyBags[bag] = false
end

function Considerater:SaveItems()
  for bag = 0,self:GetMaxBagNum() do
    self:SaveItemsInBag(bag)
  end
end

function Considerater:ResetProfile(name)
  self.db.profile.profiles[name] = {}
  for defName,defP in pairs(defaultProfiles) do
    if defName == name then
      self.db.profile.profiles[name] = self:DeepCopy(defP)
      return
    end
  end
end

function Considerater:ResetProfiles()
  self.db.profile.profiles = defaultProfiles
  self.db.char.activeProfiles = defaultActiveProfiles
end

function Considerater:GetItemEquipLoc(item)
  local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(item)
  return itemEquipLoc
end

function Considerater:GetItemTexture(item)
  local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(item)
  return itemTexture
end

function Considerater:GetMaxBagNum()
  -- Map inventory (equipped stuff) to 0
  -- Map blizzard bag -1 "main bank slots" to 1
  -- Map blizzard bags 0-4 "carried bags" to 2-6
  -- Map blizzard bags 5-11 "bank bags" to 7-13
  return 13
end

function Considerater:ToVirtualBag(bag)
  if bag == -1 then
    return 1
  else 
    return (bag + 2)
  end
end

function Considerater:GetMaxSlotNum(i)
  if i == 0 then
    return 18
  elseif i == 1 then
    return GetContainerNumSlots(-1)
  elseif i >= 2 and i <= 13 then
    return GetContainerNumSlots(i - 2)
  else
    return 0
  end
end

function Considerater:GetBagSlotItemLink(i,slot)
  if i == 0 then
    return GetInventoryItemLink("player",slot)
  elseif i == 1 then
    return GetContainerItemLink(-1,slot)
  elseif i >= 2 and i <= 13 then
    return GetContainerItemLink(i - 2,slot)
  else
    return nil
  end
end

function Considerater:MarkBlizBagDirty(bag)
  self.dirtyBags[self:ToVirtualBag(bag)] = true
  getglobal("ConsideraterUpdateFrame"):Show()
end

function Considerater:MarkInvDirty()
  self.dirtyBags[0] = true
  getglobal("ConsideraterUpdateFrame"):Show()
end

-- Gem functions


function Considerater:RefreshGemProfiles()
  local topScores = {}
  if not self.db.char.gemProfiles then
    self.db.char.gemProfiles = {}
  end

  for color, ids in pairs(self.gemTable) do
    for _,id in pairs(ids) do
      for _,profile in pairs(self.db.char.activeProfiles) do
        local gemSum = StatLogic:GetSum(id)
        if gemSum then
          local gemScore = self:ScoreItem(gemSum["link"], profile)
          if not topScores[profile] then
            topScores[profile] = {}
          end
          if not topScores[profile][color] 
             or topScores[profile][color]["score"] < gemScore then
            topScores[profile][color] = {
              ["score"] = gemScore,
              ["link"] = gemSum["link"],
              ["id"] = id,
            }
          end
        end
      end
    end
  end

  self.db.char.gemProfiles = topScores
  return topScores
end

-- Normalize gems meaning take an item and insert best-by-profile gems to avoid
-- biasing ranking because of poor gem choice.
function Considerater:NormalizeGems(item, profile)
  if(type(item) == "number") then
    local sum = StatLogic:GetSum(item)
    if not sum then
      return item
    else
      item = sum["link"]
    end
  end

  local strippedItem = StatLogic:RemoveEnchantGem(item)
  if not strippedItem then
    return item
  end

  if not self.db.char.gemProfiles or not self.db.char.gemProfiles[profile] then
    return strippedItem
  end

  local red = self.db.char.gemProfiles[profile]["red"]["link"]
  local blue = self.db.char.gemProfiles[profile]["blue"]["link"]
  local yellow = self.db.char.gemProfiles[profile]["yellow"]["link"]
  local meta = self.db.char.gemProfiles[profile]["meta"]["link"]

  if red then red = StatLogic:GetGemID(red) end
  if blue then blue = StatLogic:GetGemID(blue) end
  if yellow then yellow = StatLogic:GetGemID(yellow) end
  if meta then meta = StatLogic:GetGemID(meta) end

  local regemmedItem = 
    StatLogic:BuildGemmedTooltip(strippedItem, red, yellow, blue, meta)
  if not regemmedItem then
    return strippedItem
  else
    return regemmedItem
  end
end

function Considerater:GetNormalizedSum(item, profile)
  local strippedItem = StatLogic:RemoveEnchantGem(item)
  local sum = StatLogic:GetSum(strippedItem)
  if (sum["EMPTY_SOCKET_RED"] or
      sum["EMPTY_SOCKET_YELLOW"] or
      sum["EMPTY_SOCKET_BLUE"] or
      sum["EMPTY_SOCKET_META"]) then
    sum = StatLogic:GetSum(self:NormalizeGems(strippedItem, profile))
  end
  return sum
end

-- Consider UI

function Considerater:ConsiderUI()
  -- local gui = LibStub("AceGUI-3.0")
  -- self.conFrame = gui:Create("Frame")
  -- self.conFrame:SetTitle("Considerater")

  ConsideraterFrame:Show()
end

function Considerater_ConsiderUI_OnShow()
  self = Considerater
  if not self.uiSortByProfile or not self.uiFilterByLoc then
    self.uiSortByProfile = self.db.char.activeProfiles[1]
    self.uiFilterByLoc = "INVTYPE_HEAD"
    self:UIRefreshItemList()
  end
end

function member(x,list) 
  if not list then
    return false
  end
  for _,n in pairs(list) do
    if x == n then
      return true
    end
  end
  return false
end

function Considerater:UIRefreshItemList()
  self.uiItemList = {}
  -- Inventory, bags, bank
  for i,bag in pairs(self.db.char.cachedItems) do
    for j,slot in pairs(bag) do
      if not self.uiFilterByLoc 
         or slot["loc"] == self.uiFilterByLoc 
         or member(slot["loc"],equipLocAliases[self.uiFilterByLoc]) then
        local item = self:DeepCopy(slot)
        if i == 0 then
          item["origin"] = "Worn"
        elseif i == 1 or (i >= 7 and i <= 13) then
          item["origin"] = "Bank"
        elseif i >= 2 and i <= 6 then
          item["origin"] = "Carried"
        else
          item["origin"] = ""
        end
        table.insert(self.uiItemList, item)
      end
    end
  end

  -- Looted items
  if self.otherItemsList then
    for _,slot in pairs(self.otherItemsList) do
      if not self.uiFilterByLoc 
         or slot["loc"] == self.uiFilterByLoc 
         or member(slot["loc"],equipLocAliases[self.uiFilterByLoc]) then
        local item = self:DeepCopy(slot)
        item["origin"] = slot["origin"]
        table.insert(self.uiItemList, item)
      end
    end
  end
 
  self:UISortItemList()
  Considerater_ConsiderUI_ScrollbarUpdate()
end

function Considerater:UISortItemList()
  table.sort(self.uiItemList, function (a,b)
    return (a["scores"][self.uiSortByProfile] > b["scores"][self.uiSortByProfile])
  end)
end

function Considerater_ConsiderUI_ScrollbarUpdate()
  self = Considerater
  FauxScrollFrame_Update(ConsideraterFrame_ScrollFrame, #self.uiItemList, 10, 30)
  for line=1,10 do
    local offset = line + FauxScrollFrame_GetOffset(ConsideraterFrame_ScrollFrame)
    if offset <= #self.uiItemList then
      getglobal("ConsideraterFrame_ScrollItem"..line.."_Name"):SetText(self.uiItemList[offset]["link"])
      getglobal("ConsideraterFrame_ScrollItem"..line.."_Score"):SetText("Score: "..self.uiItemList[offset]["scores"][self.uiSortByProfile])
      getglobal("ConsideraterFrame_ScrollItem"..line.."_Origin"):SetText(self.uiItemList[offset]["origin"])
      getglobal("ConsideraterFrame_ScrollItem"..line..""):Show()
    else
      getglobal("ConsideraterFrame_ScrollItem"..line..""):Hide()
    end
  end
end

function Considerater_ConsiderUI_OnEnter(index)
  self = Considerater
  if self.uiItemList[index] then
    GameTooltip:SetOwner(ConsideraterFrame, "ANCHOR_RIGHT")
    GameTooltip:SetHyperlink(self.uiItemList[index]["link"])
  end
end

function Considerater_ConsiderUI_OnClick(index)
  self = Considerater
  if self.uiItemList[index] and IsModifiedClick() then
    -- Some mysterious Blizzard API function which appears to handle shift/ctrl
    -- clicking an item.  Awesome.
    HandleModifiedItemClick(self.uiItemList[index]["link"])
  end
end

function Considerater_ConsiderUI_ProfileDropDown_Initialize()
  self = Considerater
  for i,prof in pairs(self.db.char.activeProfiles) do
    info = UIDropDownMenu_CreateInfo() 
    info.text = prof
    info.value = prof
    info.func = function (dropDown) 
      self.uiSortByProfile = dropDown.value
      UIDropDownMenu_SetSelectedValue(ConsideraterFrame_ProfileDropDown, dropDown.value)
      self:UISortItemList()
      Considerater_ConsiderUI_ScrollbarUpdate()
    end
    UIDropDownMenu_AddButton(info)
  end
end

function Considerater_ConsiderUI_LocDropDown_Initialize()
  self = Considerater
  for _,loc in pairs(equipLocs) do
    info = UIDropDownMenu_CreateInfo() 
    info.text = loc["text"]
    info.value = loc["value"]
    info.func = function (dropDown) 
      self.uiFilterByLoc = dropDown.value
      UIDropDownMenu_SetSelectedValue(ConsideraterFrame_LocDropDown, dropDown.value)
      self:UIRefreshItemList()
      Considerater_ConsiderUI_ScrollbarUpdate()
    end
    UIDropDownMenu_AddButton(info)
  end
end

-- Configuration UI stuff

function Considerater:GetProfileNames()
  local names = {}
  for n,_ in pairs(self.db.profile.profiles) do
    table.insert(names, n)
  end
  return names
end

function Considerater:IsProfileActive(p)
  for i,n in pairs(self.db.char.activeProfiles) do
    if n == p then
      return i
    end
  end
  return false
end

function Considerater:SetProfileActive(p,val)
  local i = self:IsProfileActive(p)
  if val and not i then
    table.insert(self.db.char.activeProfiles, p)
  elseif not val and i then
    table.remove(self.db.char.activeProfiles, i)
  end
end

function Considerater:EqualItemLinks(l1, l2)
  local name1 = GetItemInfo(l1)
  local name2 = GetItemInfo(l2)
  return (name1 == name2)
end

function Considerater:DeepCopy(t)
  if type(t) == "table" then
    local s = {}
    for k,v in pairs(t) do
      s[k] = self:DeepCopy(v)
    end
    return s
  else
    return t
  end
end

function Considerater:DeepEqual(t1,t2)
  if type(t1) ~= type(t2) then
    return false
  elseif type(t1) == "table" then
    for k,v in pairs(t1) do
      if not self:DeepEqual(v,t2[k]) then
        return false
      end
    end
    return true
  else
    return (t1 == t2)
  end
end

